#ifndef MSP_GEOMETRY_COMPOSITESHAPE_H_
#define MSP_GEOMETRY_COMPOSITESHAPE_H_

#include <algorithm>
#include <stdexcept>
#include <vector>
#include "shape.h"

namespace Msp {
namespace Geometry {

/**
Common operations for shapes composed of other shapes.
*/
template<typename T, unsigned D, typename O>
class CompositeShape: public Shape<T, D>
{
protected:
	typedef O Ops;
	typedef std::vector<Shape<T, D> *> ShapeArray;

	ShapeArray shapes;
	unsigned max_isect;
	unsigned min_scratch;

	CompositeShape() { }
	CompositeShape(const Shape<T, D> &, const Shape<T, D> &);
	template<typename Iter>
	void init_from_iter_range(const Iter &, const Iter &);
private:
	void init();
protected:
	CompositeShape(const CompositeShape &);
	CompositeShape &operator=(const CompositeShape &);
public:
	virtual ~CompositeShape();

	virtual BoundingBox<T, D> get_axis_aligned_bounding_box() const;
	virtual bool contains(const LinAl::Vector<T, D> &) const;
	virtual unsigned get_max_ray_intersections() const { return max_isect; }
	virtual unsigned get_intersections(const Ray<T, D> &, SurfacePoint<T, D> *, unsigned) const;
};

template<typename T, unsigned D, typename O>
inline CompositeShape<T, D, O>::CompositeShape(const Shape<T, D> &s1, const Shape<T, D> &s2)
{
	shapes.reserve(2);
	shapes.push_back(s1.clone());
	shapes.push_back(s2.clone());
	init();
}

template<typename T, unsigned D, typename O>
template<typename Iter>
inline void CompositeShape<T, D, O>::init_from_iter_range(const Iter &begin, const Iter &end)
{
	if(begin==end)
		throw std::invalid_argument("CompositeShape::init_from_iter_range");

	for(Iter i=begin; i!=end; ++i)
		shapes.push_back((*i)->clone());
	init();
}

template<typename T, unsigned D, typename O>
inline void CompositeShape<T, D, O>::init()
{
	max_isect = 0;
	min_scratch = 0;
	for(typename ShapeArray::const_iterator i=shapes.begin(); i!=shapes.end(); ++i)
	{
		unsigned mi = (*i)->get_max_ray_intersections();
		max_isect += mi;
		min_scratch = std::max(min_scratch, mi);
	}
}

template<typename T, unsigned D, typename O>
inline CompositeShape<T, D, O>::CompositeShape(const CompositeShape<T, D, O> &other):
	shapes(other.shapes),
	max_isect(other.max_isect),
	min_scratch(other.min_scratch)
{
	for(typename ShapeArray::iterator i=shapes.begin(); i!=shapes.end(); ++i)
		*i = (*i)->clone();
}

template<typename T, unsigned D, typename O>
inline CompositeShape<T, D, O> &CompositeShape<T, D, O>::operator=(const CompositeShape<T, D, O> &other)
{
	for(typename ShapeArray::iterator i=shapes.begin(); i!=shapes.end(); ++i)
		delete *i;

	shapes = other.shapes;
	for(typename ShapeArray::iterator i=shapes.begin(); i!=shapes.end(); ++i)
		*i = (*i)->clone();

	max_isect = other.max_isect;
	min_scratch = other.min_scratch;
}

template<typename T, unsigned D, typename O>
inline CompositeShape<T, D, O>::~CompositeShape()
{
	for(typename ShapeArray::iterator i=shapes.begin(); i!=shapes.end(); ++i)
		delete *i;
}

template<typename T, unsigned D, typename O>
inline BoundingBox<T, D> CompositeShape<T, D, O>::get_axis_aligned_bounding_box() const
{
	BoundingBox<T, D> aabb;
	for(typename ShapeArray::const_iterator i=shapes.begin(); i!=shapes.end(); ++i)
	{
		if(i==shapes.begin())
			aabb = (*i)->get_axis_aligned_bounding_box();
		else
			aabb = Ops::combine_aabb(aabb, (*i)->get_axis_aligned_bounding_box());
	}
	return aabb;
}

template<typename T, unsigned D, typename O>
inline bool CompositeShape<T, D, O>::contains(const LinAl::Vector<T, D> &point) const
{
	bool inside = Ops::init_inside();
	for(typename ShapeArray::const_iterator i=shapes.begin(); i!=shapes.end(); ++i)
	{
		inside = Ops::combine_inside(inside, (*i)->contains(point));
		if(Ops::is_inside_decided(inside))
			break;
	}
	return inside;
}

template<typename T, unsigned D, typename O>
inline unsigned CompositeShape<T, D, O>::get_intersections(const Ray<T, D> &ray, SurfacePoint<T, D> *points, unsigned size) const
{
	SurfacePoint<T, D> *buffer = points;
	unsigned buf_size = size;
	if(!points)
	{
		buffer = new SurfacePoint<T, D>[min_scratch];
		buf_size = min_scratch;
	}

	unsigned n = 0;
	for(typename ShapeArray::const_iterator i=shapes.begin(); (n<size && i!=shapes.end()); ++i)
	{
		unsigned base = (points ? n : 0);
		unsigned count = (*i)->get_intersections(ray, buffer+base, buf_size-base);
		for(unsigned j=0; (n<size && j<count); ++j)
		{
			SurfacePoint<T, D> &pt = buffer[base+j];

			bool surface = Ops::init_surface();
			for(typename ShapeArray::const_iterator k=shapes.begin(); k!=shapes.end(); ++k)
				if(k!=i)
					surface = Ops::combine_surface(surface, (*k)->contains(pt.position));

			if(surface)
			{
				if(points && base+j!=n)
					points[n] = pt;

				++n;
			}
		}
	}

	if(points)
		sort_points(points, n);
	else
		delete[] buffer;

	return n;
}

} // namespace Geometry
} // namespace Msp

#endif
