#ifndef MSP_LINAL_MATRIX_H_
#define MSP_LINAL_MATRIX_H_

#include <algorithm>
#include "vector.h"

namespace Msp {
namespace LinAl {

/**
A general mathematical matrix with M rows and N columns.
*/
template<typename T, unsigned M, unsigned N>
class Matrix
{
private:
	T data[M*N];

public:
	Matrix();
	Matrix(const T *);
	template<typename U>
	Matrix(const Matrix<U, M, N> &);

	static Matrix from_columns(const Vector<T, M> *);
	static Matrix from_rows(const Vector<T, N> *);

	T &element(unsigned i, unsigned j) { return data[i+M*j]; }
	const T &element(unsigned i, unsigned j) const { return data[i+M*j]; }
	T &operator()(unsigned i, unsigned j) { return element(i, j); }
	const T &operator()(unsigned i, unsigned j) const { return element(i, j); }

	Matrix &operator*=(T);
	Matrix &operator/=(T);
	Matrix &operator+=(const Matrix &);
	Matrix &operator-=(const Matrix &);

	Matrix &exchange_rows(unsigned, unsigned);
	Matrix &multiply_row(unsigned, T);
	Matrix &add_row(unsigned, unsigned, T);
};

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N>::Matrix()
{
	std::fill(data, data+M*N, T());
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N>::Matrix(const T *d)
{
	std::copy(d, d+M*N, data);
}

template<typename T, unsigned M, unsigned N>
template<typename U>
inline Matrix<T, M, N>::Matrix(const Matrix<U, M, N> &other)
{
	for(unsigned i=0; i<M; ++i)
		for(unsigned j=0; j<N; ++j)
			element(i, j) = other(i, j);
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N> Matrix<T, M, N>::from_columns(const Vector<T, M> *v)
{
	Matrix<T, M, N> m;
	for(unsigned i=0; i<M; ++i)
		for(unsigned j=0; j<N; ++j)
			m(i, j) = v[j][i];
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N> Matrix<T, M, N>::from_rows(const Vector<T, N> *v)
{
	Matrix<T, M, N> m;
	for(unsigned i=0; i<M; ++i)
		for(unsigned j=0; j<N; ++j)
			m(i, j) = v[i][j];
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N> &Matrix<T, M, N>::operator*=(T s)
{
	for(unsigned i=0; i<M*N; ++i)
		data[i] *= s;
	return *this;
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N> operator*(const Matrix<T, M, N> &m, T s)
{
	Matrix<T, M, N> r(m);
	return r *= s;
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N> operator*(T s, const Matrix<T, M, N> &m)
{
	return m*s;
}

template<typename T, unsigned M, unsigned P, unsigned N>
inline Matrix<T, M, N> operator*(const Matrix<T, M, P> &m1, const Matrix<T, P, N> &m2)
{
	Matrix<T, M, N> r;
	for(unsigned i=0; i<M; ++i)
		for(unsigned j=0; j<N; ++j)
			for(unsigned k=0; k<P; ++k)
				r(i, j) += m1(i, k)*m2(k, j);
	return r;
}

template<typename T, unsigned M, unsigned N>
inline Vector<T, M> operator*(const Matrix<T, M, N> &m, const Vector<T, N> &v)
{
	Vector<T, M> r;
	for(unsigned i=0; i<M; ++i)
		for(unsigned j=0; j<N; ++j)
			r[i] += m(i, j)*v[j];
	return r;
}

template<typename T, unsigned M, unsigned N>
inline Vector<T, N> operator*(const Vector<T, M> &v, const Matrix<T, M, N> &m)
{
	Vector<T, N> r;
	for(unsigned j=0; j<N; ++j)
		for(unsigned i=0; i<M; ++i)
			r[j] += v[i]*m(i, j);
	return r;
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N> &Matrix<T, M, N>::operator/=(T s)
{
	for(unsigned i=0; i<M*N; ++i)
		data[i] /= s;
	return *this;
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N> operator/(const Matrix<T, M, N> &m, T s)
{
	Matrix<T, M, N> r(m);
	return r /= s;
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N> &Matrix<T, M, N>::operator+=(const Matrix<T, M, N> &m)
{
	for(unsigned i=0; i<M*N; ++i)
		data[i] += m.data[i];
	return *this;
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N> operator+(const Matrix<T, M, N> &m1, const Matrix<T, M, N> &m2)
{
	Matrix<T, M, N> r(m1);
	return r += m2;
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N> &Matrix<T, M, N>::operator-=(const Matrix<T, M, N> &m)
{
	for(unsigned i=0; i<M*N; ++i)
		data[i] -= m.data[i];
	return *this;
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N> operator-(const Matrix<T, M, N> &m1, const Matrix<T, M, N> &m2)
{
	Matrix<T, M, N> r(m1);
	return r -= m2;
}

template<typename T, unsigned M, unsigned N>
inline bool operator==(const Matrix<T, M, N> &a, const Matrix<T, M, N> &b)
{
	for(unsigned j=0; j<N; ++j)
		for(unsigned i=0; i<M; ++i)
			if(a(i, j)!=b(i, j))
				return false;
	return true;
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N> &Matrix<T, M, N>::exchange_rows(unsigned i, unsigned j)
{
	using std::swap;
	for(unsigned k=0; k<N; ++k)
		swap(element(i, k), element(j, k));
	return *this;
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N> &Matrix<T, M, N>::multiply_row(unsigned i, T s)
{
	for(unsigned k=0; k<N; ++k)
		element(i, k) *= s;
	return *this;
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, M, N> &Matrix<T, M, N>::add_row(unsigned i, unsigned j, T s)
{
	for(unsigned k=0; k<N; ++k)
		element(j, k) += element(i, k)*s;
	return *this;
}

template<typename T, unsigned M, unsigned N>
inline Matrix<T, N, M> transpose(const Matrix<T, M, N> &m)
{
	Matrix<T, N, M> r;
	for(unsigned j=0; j<N; ++j)
		for(unsigned i=0; i<M; ++i)
			r(j, i) = m(i, j);
	return r;
}

} // namespace LinAl
} // namespace Msp

#endif
